/*
 * ============================================================================
 * Licensed Materials - Property of IBM
 * Project  Zero
 *
 * (C) Copyright IBM Corp. 2007  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 * ============================================================================
 */

package zero.resource.backends.sql

import zero.resource.Collection
import zero.resource.Type
import zero.resource.TypeCollection
import zero.data.handler.MapKeyedResultHandler
import zero.core.context.GlobalContext
import zero.core.context.GlobalContextURIs.*
import zero.json.Json
import java.sql.SQLException
import com.ibm.pdq.runtime.exception.DataRuntimeException
import zero.resource.exceptions.DataParseException

class Sql {
    
    static load_type_data(type_name, fixtures) {
        def dependencies = GlobalContext.zget(Config.dependencies)
        dependencies.each { dependency ->
            def dependency_root = GlobalContext.zget("${Config.dependencies}/${dependency}")
            def fixture_root = new File("${dependency_root}/app/models/fixtures/")
            if (fixture_root.exists()) {
                fixtures.each { fixture ->
                    def fixture_file = new File("${dependency_root}/app/models/fixtures/${fixture}")
                    try {
                        def json = Json.decode(fixture_file.text)
                        def rows = json.findAll { it.type==type_name }
                        rows.each { member ->
                            def type = TypeCollection.retrieve(member.type)
                            type.create(member.fields)
                        }
                    } catch (Exception e) {
                        throw new DataParseException("error parsing ${fixture_file.canonicalPath}: ${e.message}", e)
                    }
                }
            }
        }
    }
    
    static load_data(fixtures) {
        // fixtures is array of strings
        def dependencies = GlobalContext.zget(Config.dependencies)
        dependencies.each { dependency ->
            def dependency_root = GlobalContext.zget("${Config.dependencies}/${dependency}")
            def fixture_root = new File("${dependency_root}/app/models/fixtures/")
            if (fixture_root.exists()) {
                fixtures.each { fixture ->
                    def fixture_file = new File("${dependency_root}/app/models/fixtures/${fixture}")
                    try {
                        def json = Json.decode(fixture_file.text)
                        json.each { member ->
                            def type = TypeCollection.retrieve(member.type)
                            type.create(member.fields)
                        }
                    } catch (Exception e) {
                        throw new DataParseException("error parsing ${fixture_file.canonicalPath}: ${e.message}", e)
                    }
                }
            }
        }
    }
    
    static reset_data(fixtures) {
        def connection = Collection.data.retrieveConnection()
        def statements = Sql.sql_reset()
            
        statements.each { statement ->
            try {
                connection.prepareStatement(statement).execute()
            } catch (SQLException e) {
                if (!statement.trim().startsWith('DROP')) {
                    throw new DataRuntimeException(e)
                }
            }
        }
        
        load_data(fixtures)
    }
    
    static load_initial_data() {
        load_data(['initial_data.json'])
    }
    
    static reset_initial_data() {
        reset_data(['initial_data.json'])
    }
    
    static sync_db() {
        // Get a list of all existing database tables, so we know what needs to
        // be added.

        def connection = Collection.data.retrieveConnection()
        def filter = new String[1];filter[0] = 'TABLE'
        def tables_result_set = connection.metaData.getTables(null,null,null,filter)
        def tables = [] + (new MapKeyedResultHandler('TABLE_NAME').handle(tables_result_set)).keySet()

        // Create the tables for each model
        for (type in TypeCollection.list()) {
            def raw_table_name = type.dbTable.toUpperCase()
            if (!tables.contains(raw_table_name)) {
                def sql = sql_type_create(type)
                try {
                    connection.prepareStatement(sql).execute()
                    load_type_data(type.typeName, ['initial_data.json'])
                    tables << raw_table_name
                } catch (SQLException e) {
                    if (!sql.trim().startsWith('DROP'))
                        throw new DataRuntimeException(e)
                }
                
            }
        }

    }
    
    static sql_all() {
        return sql_create() // + sql_custom() + sql_indexes()
    }
    
    static sql_reset() {
        return sql_delete() + sql_all()
    }
    
    static sql_create() {
        def statements = [:]
        TypeCollection.list().each { type ->
            def raw_table_name = type.dbTable.toUpperCase()
            if (!statements.containsKey(raw_table_name)) {
                statements[raw_table_name] = sql_type_create(type)
            }
        }
        return statements.values().asList()
    }
    
    static sql_type_create(type) {
        def table_output = []
        def meta_fields = Type.meta_fields
        
        type.fields.each { field_name, field_spec ->
            // only do non-meta fields
            if (!meta_fields.contains(field_name)) {
                def field_output = [field_spec.column_name, field_spec.dbType()]
                if (field_spec.required) {
                    field_output << 'NOT NULL'
                }
                table_output << field_output.join(' ')
            }
        }
        
        // force meta fields to be NOT NULL since they are not required for
        // validation but required for persistence
        meta_fields.each { field_name ->
            def field_spec = type.fields[field_name]
            table_output << [field_spec.column_name,field_spec.dbType(),'NOT NULL'].join(' ')
        }
        
        // build CREATE TABLE statement
        def full_statement = ["CREATE TABLE $type.dbTable ("]
        table_output.eachWithIndex { line, i ->
            def line_end = i<table_output.size()-1?',':''
            full_statement << "    ${line}${line_end}"
        }
        full_statement << ')'
        return full_statement.join('\n')
    }
    
    static sql_delete() {
        def output = []
        // because we don't support relationships, don't concern about ordering
        // at present
        
        for (type in TypeCollection.list()) {
            def raw_table_name = type.dbTable
            output << "DROP TABLE ${raw_table_name}".toString()
        }
        return output
    }

    static installed_types(table_list) {
        return TypeCollection.list().findAll { type ->
            table_list.contains(type.dbTable.toUpperCase())
        }
    }
}